﻿using NtApiDotNet;
using System;
using System.IO;
using System.Runtime.InteropServices;
using System.Runtime.InteropServices.ComTypes;

namespace POC
{
    [Flags]
    public enum STGM
    {
        READ = 0x00000000,
        WRITE = 0x00000001,
        READWRITE = 0x00000002,
        SHARE_DENY_NONE = 0x00000040,
        SHARE_DENY_READ = 0x00000030,
        SHARE_DENY_WRITE = 0x00000020,
        SHARE_EXCLUSIVE = 0x00000010,
        PRIORITY = 0x00040000,
        CREATE = 0x00001000,
        CONVERT = 0x00020000,
        FAILIFTHERE = READ,
        DIRECT = READ,
        TRANSACTED = 0x00010000,
        NOSCRATCH = 0x00100000,
        NOSNAPSHOT = 0x00200000,
        SIMPLE = 0x08000000,
        DIRECT_SWMR = 0x00400000,
        DELETEONRELEASE = 0x04000000
    }

    sealed class StreamWrapper : Stream
    {
        private IStream _stm;

        public StreamWrapper(IStream stm)
        {
            _stm = stm;
        }

        private System.Runtime.InteropServices.ComTypes.STATSTG Stat()
        {
            _stm.Stat(out System.Runtime.InteropServices.ComTypes.STATSTG stat, 1);
            return stat;
        }

        protected override void Dispose(bool disposing)
        {
            Marshal.FinalReleaseComObject(_stm);
            base.Dispose(disposing);
        }

        public override bool CanRead => true;

        public override bool CanSeek => true;

        public override bool CanWrite
        {
            get
            {
                STGM mode = (STGM)Stat().grfMode;
                return ((mode & STGM.WRITE) == STGM.WRITE) || ((mode & STGM.READWRITE) == STGM.READWRITE);
            }
        }

        public override long Length => Stat().cbSize;

        public override long Position
        {
            get
            {
                return Seek(0, SeekOrigin.Current);
            }

            set
            {
                // STREAM_SEEK_SET == 0
                _stm.Seek(value, 0, IntPtr.Zero);
            }
        }

        public override void Flush()
        {
            _stm.Commit(0);
        }

        public override int Read(byte[] buffer, int offset, int count)
        {
            if (offset == 0)
            {
                using (var len = new SafeStructureInOutBuffer<int>())
                {
                    _stm.Read(buffer, count, len.DangerousGetHandle());
                    return len.Result;
                }
            }
            else
            {
                using (var len = new SafeStructureInOutBuffer<int>())
                {
                    byte[] temp_buffer = new byte[count];
                    _stm.Read(temp_buffer, count, len.DangerousGetHandle());
                    int read_len = len.Result;
                    Buffer.BlockCopy(temp_buffer, 0, buffer, offset, count);
                    return read_len;
                }
            }
        }

        public override long Seek(long offset, SeekOrigin origin)
        {
            using (var buffer = new SafeStructureInOutBuffer<long>())
            {
                _stm.Seek(0, (int)origin, buffer.DangerousGetHandle());
                return buffer.Result;
            }
        }

        public override void SetLength(long value)
        {
            _stm.SetSize(value);
        }

        public override void Write(byte[] buffer, int offset, int count)
        {
            _stm.Write(buffer, count, IntPtr.Zero);
        }
    }
}
